import argparse
from pathlib import Path
from uuid import uuid4

import optuna
import torch
from optuna.samplers import TPESampler

import diart.argdoc as argdoc
from diart.inference import Benchmark
from diart.optim import Optimizer, HyperParameter
from diart.pipelines import PipelineConfig


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("root", type=str, help="Directory with audio files CONVERSATION.(wav|flac|m4a|...)")
    parser.add_argument("--reference", required=True, type=str,
                        help="Directory with RTTM files CONVERSATION.rttm. Names must match audio files")
    parser.add_argument("--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5")
    parser.add_argument("--latency", default=0.5, type=float, help=f"{argdoc.LATENCY}. Defaults to 0.5")
    parser.add_argument("--tau", default=0.5, type=float, help=f"{argdoc.TAU}. Defaults to 0.5")
    parser.add_argument("--rho", default=0.3, type=float, help=f"{argdoc.RHO}. Defaults to 0.3")
    parser.add_argument("--delta", default=1, type=float, help=f"{argdoc.DELTA}. Defaults to 1")
    parser.add_argument("--gamma", default=3, type=float, help=f"{argdoc.GAMMA}. Defaults to 3")
    parser.add_argument("--beta", default=10, type=float, help=f"{argdoc.BETA}. Defaults to 10")
    parser.add_argument("--max-speakers", default=20, type=int, help=f"{argdoc.MAX_SPEAKERS}. Defaults to 20")
    parser.add_argument("--batch-size", default=32, type=int, help=f"{argdoc.BATCH_SIZE}. Defaults to 32")
    parser.add_argument("--cpu", dest="cpu", action="store_true",
                        help=f"{argdoc.CPU}. Defaults to GPU if available, CPU otherwise")
    parser.add_argument("--hparams", nargs="+", default=("tau_active", "rho_update", "delta_new"),
                        help="Hyper-parameters to optimize. Must match names in `PipelineConfig`. Defaults to tau_active, rho_update and delta_new")
    parser.add_argument("--num-iter", default=100, type=int, help="Number of optimization trials")
    parser.add_argument("--storage", type=str,
                        help="Optuna storage string. If provided, continue a previous study instead of creating one. The database name must match the study name")
    parser.add_argument("--output", required=True, type=str, help="Working directory")
    args = parser.parse_args()
    args.output = Path(args.output)
    args.output.mkdir(parents=True, exist_ok=True)
    args.device = torch.device("cpu") if args.cpu else None

    # Assign unique worker ID
    idx = uuid4()

    # Create benchmark object to run the pipeline on a set of files
    work_path = args.output / f"worker-{idx}"
    benchmark = Benchmark(
        args.root,
        args.reference,
        work_path,
        show_progress=True,
        show_report=False,
        batch_size=args.batch_size
    )

    # Create the base configuration for each trial
    base_config = PipelineConfig.from_namespace(args)

    # Create hyper-parameters to optimize
    hparams = [HyperParameter.from_name(name) for name in args.hparams]

    # Use a custom storage if given
    study_or_path = args.output
    if args.storage is not None:
        db_name = Path(args.storage).stem
        study_or_path = optuna.load_study(db_name, args.storage, TPESampler())

    # Run optimization
    optimizer = Optimizer(benchmark, base_config, hparams, study_or_path)
    optimizer.optimize(num_iter=args.num_iter, show_progress=True)

    # Clean temporary directory
    work_path.rmdir()


if __name__ == "__main__":
    run()



from typing import Union, Optional

import numpy as np
import torch
from pyannote.core import SlidingWindow, SlidingWindowFeature

TemporalFeatures = Union[SlidingWindowFeature, np.ndarray, torch.Tensor]


class TemporalFeatureFormatterState:
    """
    Represents the recorded type of a temporal feature formatter.
    Its job is to transform temporal features into tensors and
    recover the original format on other features.
    """
    def to_tensor(self, features: TemporalFeatures) -> torch.Tensor:
        raise NotImplementedError

    def to_internal_type(self, features: torch.Tensor) -> TemporalFeatures:
        """
        Cast `features` to the representing type and remove batch dimension if required.

        Parameters
        ----------
        features: torch.Tensor, shape (batch, frames, dim)
            Batched temporal features.
        Returns
        -------
        new_features: SlidingWindowFeature or numpy.ndarray or torch.Tensor, shape (batch, frames, dim)
        """
        raise NotImplementedError


class SlidingWindowFeatureFormatterState(TemporalFeatureFormatterState):
    def __init__(self, duration: float):
        self.duration = duration
        self._cur_start_time = 0

    def to_tensor(self, features: SlidingWindowFeature) -> torch.Tensor:
        msg = "Features sliding window duration and step must be equal"
        assert features.sliding_window.duration == features.sliding_window.step, msg
        self._cur_start_time = features.sliding_window.start
        return torch.from_numpy(features.data)

    def to_internal_type(self, features: torch.Tensor) -> TemporalFeatures:
        batch_size, num_frames, _ = features.shape
        assert batch_size == 1, "Batched SlidingWindowFeature objects are not supported"
        # Calculate resolution
        resolution = self.duration / num_frames
        # Temporal shift to keep track of current start time
        resolution = SlidingWindow(start=self._cur_start_time, duration=resolution, step=resolution)
        return SlidingWindowFeature(features.squeeze(dim=0).cpu().numpy(), resolution)


class NumpyArrayFormatterState(TemporalFeatureFormatterState):
    def to_tensor(self, features: np.ndarray) -> torch.Tensor:
        return torch.from_numpy(features)

    def to_internal_type(self, features: torch.Tensor) -> TemporalFeatures:
        return features.cpu().numpy()


class PytorchTensorFormatterState(TemporalFeatureFormatterState):
    def to_tensor(self, features: torch.Tensor) -> torch.Tensor:
        return features

    def to_internal_type(self, features: torch.Tensor) -> TemporalFeatures:
        return features


class TemporalFeatureFormatter:
    """
    Manages the typing and format of temporal features.
    When casting temporal features as torch.Tensor, it remembers its
    type and format so it can lately restore it on other temporal features.
    """
    def __init__(self):
        self.state: Optional[TemporalFeatureFormatterState] = None

    def set_state(self, features: TemporalFeatures):
        if self.state is not None:
            return

        if isinstance(features, SlidingWindowFeature):
            msg = "Features sliding window duration and step must be equal"
            assert features.sliding_window.duration == features.sliding_window.step, msg
            self.state = SlidingWindowFeatureFormatterState(
                features.data.shape[0] * features.sliding_window.duration,
            )
        elif isinstance(features, np.ndarray):
            self.state = NumpyArrayFormatterState()
        elif isinstance(features, torch.Tensor):
            self.state = PytorchTensorFormatterState()
        else:
            msg = "Unknown format. Provide one of SlidingWindowFeature, numpy.ndarray, torch.Tensor"
            raise ValueError(msg)

    def cast(self, features: TemporalFeatures) -> torch.Tensor:
        """
        Transform features into a `torch.Tensor` and add batch dimension if missing.

        Parameters
        ----------
        features: SlidingWindowFeature or numpy.ndarray or torch.Tensor
            Shape (frames, dim) or (batch, frames, dim)

        Returns
        -------
        features: torch.Tensor, shape (batch, frames, dim)
        """
        # Set state if not initialized
        self.set_state(features)
        # Convert features to tensor
        data = self.state.to_tensor(features)
        # Make sure there's a batch dimension
        msg = "Temporal features must be 2D or 3D"
        assert data.ndim in (2, 3), msg
        if data.ndim == 2:
            data = data.unsqueeze(0)
        return data.float()

    def restore_type(self, features: torch.Tensor) -> TemporalFeatures:
        """
        Cast `features` to the internal type and remove batch dimension if required.

        Parameters
        ----------
        features: torch.Tensor, shape (batch, frames, dim)
            Batched temporal features.
        Returns
        -------
        new_features: SlidingWindowFeature or numpy.ndarray or torch.Tensor, shape (batch, frames, dim)
        """
        return self.state.to_internal_type(features)



from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Text, Optional, Union

from optuna import TrialPruned, Study, create_study
from optuna.samplers import TPESampler
from optuna.trial import Trial, FrozenTrial
from tqdm import trange, tqdm

from .audio import FilePath
from .benchmark import Benchmark
from .pipelines import PipelineConfig, OnlineSpeakerDiarization


@dataclass
class HyperParameter:
    name: Text
    low: float
    high: float

    @staticmethod
    def from_name(name: Text) -> 'HyperParameter':
        if name == "tau_active":
            return TauActive
        if name == "rho_update":
            return RhoUpdate
        if name == "delta_new":
            return DeltaNew
        raise ValueError(f"Hyper-parameter '{name}' not recognized")


TauActive = HyperParameter("tau_active", low=0, high=1)
RhoUpdate = HyperParameter("rho_update", low=0, high=1)
DeltaNew = HyperParameter("delta_new", low=0, high=2)


class Optimizer:
    def __init__(
        self,
        benchmark: Benchmark,
        base_config: PipelineConfig,
        hparams: Iterable[HyperParameter],
        study_or_path: Union[FilePath, Study],
    ):
        self.benchmark = benchmark
        self.base_config = base_config
        self.hparams = hparams
        self._progress: Optional[tqdm] = None

        if isinstance(study_or_path, Study):
            self.study = study_or_path
        elif isinstance(study_or_path, str) or isinstance(study_or_path, Path):
            self.study = create_study(
                storage="sqlite:///" + str(study_or_path / f"{study_or_path.stem}.db"),
                sampler=TPESampler(),
                study_name=study_or_path.stem,
                direction="minimize",
                load_if_exists=True,
            )
        else:
            msg = f"Expected Study object or path-like, but got {type(study_or_path).__name__}"
            raise ValueError(msg)

    @property
    def best_performance(self):
        return self.study.best_value

    @property
    def best_hparams(self):
        return self.study.best_params

    def _callback(self, study: Study, trial: FrozenTrial):
        if self._progress is None:
            return
        self._progress.update(1)
        self._progress.set_description(f"Trial {trial.number + 1}")
        values = {"best_der": study.best_value}
        for name, value in study.best_params.items():
            values[f"best_{name}"] = value
        self._progress.set_postfix(OrderedDict(values))

    def objective(self, trial: Trial) -> float:
        # Set suggested values for optimized hyper-parameters
        trial_config = vars(self.base_config)
        for hparam in self.hparams:
            trial_config[hparam.name] = trial.suggest_uniform(
                hparam.name, hparam.low, hparam.high
            )

        # Prune trial if required
        if trial.should_prune():
            raise TrialPruned()

        # Instantiate pipeline with the new configuration
        pipeline = OnlineSpeakerDiarization(PipelineConfig(**trial_config), profile=False)

        # Run pipeline over the dataset
        report = self.benchmark(pipeline)

        # Clean RTTM files
        for tmp_file in self.benchmark.output_path.iterdir():
            if tmp_file.name.endswith(".rttm"):
                tmp_file.unlink()

        # Extract DER from report
        return report.loc["TOTAL", "diarization error rate"]["%"]

    def optimize(self, num_iter: int, show_progress: bool = True):
        self._progress = None
        if show_progress:
            self._progress = trange(num_iter)
            last_trial = -1
            if self.study.trials:
                last_trial = self.study.trials[-1].number
            self._progress.set_description(f"Trial {last_trial + 1}")
        self.study.optimize(self.objective, num_iter, callbacks=[self._callback])



import argparse
from pathlib import Path

import torch

import diart.argdoc as argdoc
import diart.sources as src
from diart.inference import RealTimeInference
from diart.pipelines import OnlineSpeakerDiarization, PipelineConfig


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("source", type=str, help="Path to an audio file | 'microphone'")
    parser.add_argument("--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5")
    parser.add_argument("--latency", default=0.5, type=float, help=f"{argdoc.LATENCY}. Defaults to 0.5")
    parser.add_argument("--tau", default=0.5, type=float, help=f"{argdoc.TAU}. Defaults to 0.5")
    parser.add_argument("--rho", default=0.3, type=float, help=f"{argdoc.RHO}. Defaults to 0.3")
    parser.add_argument("--delta", default=1, type=float, help=f"{argdoc.DELTA}. Defaults to 1")
    parser.add_argument("--gamma", default=3, type=float, help=f"{argdoc.GAMMA}. Defaults to 3")
    parser.add_argument("--beta", default=10, type=float, help=f"{argdoc.BETA}. Defaults to 10")
    parser.add_argument("--max-speakers", default=20, type=int, help=f"{argdoc.MAX_SPEAKERS}. Defaults to 20")
    parser.add_argument("--no-plot", dest="no_plot", action="store_true", help="Skip plotting for faster inference")
    parser.add_argument("--cpu", dest="cpu", action="store_true",
                        help=f"{argdoc.CPU}. Defaults to GPU if available, CPU otherwise")
    parser.add_argument("--output", type=str,
                        help=f"{argdoc.OUTPUT}. Defaults to home directory if SOURCE == 'microphone' or parent directory if SOURCE is a file")
    args = parser.parse_args()
    args.device = torch.device("cpu") if args.cpu else None

    # Define online speaker diarization pipeline
    config = PipelineConfig.from_namespace(args)
    pipeline = OnlineSpeakerDiarization(config, profile=True)

    # Manage audio source
    if args.source != "microphone":
        args.source = Path(args.source).expanduser()
        args.output = args.source.parent if args.output is None else Path(args.output)
        audio_source = src.FileAudioSource(
            args.source, config.sample_rate, config.duration, config.step
        )
    else:
        args.output = Path("~/").expanduser() if args.output is None else Path(args.output)
        audio_source = src.MicrophoneAudioSource(config.sample_rate)

    # Run online inference
    RealTimeInference(args.output, do_plot=not args.no_plot)(pipeline, audio_source)


if __name__ == "__main__":
    run()



import argparse

import torch

import diart.argdoc as argdoc
from diart.inference import Benchmark
from diart.pipelines import OnlineSpeakerDiarization, PipelineConfig


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("root", type=str, help="Directory with audio files CONVERSATION.(wav|flac|m4a|...)")
    parser.add_argument("--reference", type=str,
                        help="Optional. Directory with RTTM files CONVERSATION.rttm. Names must match audio files")
    parser.add_argument("--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5")
    parser.add_argument("--latency", default=0.5, type=float, help=f"{argdoc.LATENCY}. Defaults to 0.5")
    parser.add_argument("--tau", default=0.5, type=float, help=f"{argdoc.TAU}. Defaults to 0.5")
    parser.add_argument("--rho", default=0.3, type=float, help=f"{argdoc.RHO}. Defaults to 0.3")
    parser.add_argument("--delta", default=1, type=float, help=f"{argdoc.DELTA}. Defaults to 1")
    parser.add_argument("--gamma", default=3, type=float, help=f"{argdoc.GAMMA}. Defaults to 3")
    parser.add_argument("--beta", default=10, type=float, help=f"{argdoc.BETA}. Defaults to 10")
    parser.add_argument("--max-speakers", default=20, type=int, help=f"{argdoc.MAX_SPEAKERS}. Defaults to 20")
    parser.add_argument("--batch-size", default=32, type=int, help=f"{argdoc.BATCH_SIZE}. Defaults to 32")
    parser.add_argument("--cpu", dest="cpu", action="store_true",
                        help=f"{argdoc.CPU}. Defaults to GPU if available, CPU otherwise")
    parser.add_argument("--output", type=str, help=f"{argdoc.OUTPUT}. Defaults to `root`")
    args = parser.parse_args()
    args.device = torch.device("cpu") if args.cpu else None

    benchmark = Benchmark(
        args.root,
        args.reference,
        args.output,
        show_progress=True,
        show_report=True,
        batch_size=args.batch_size
    )

    benchmark(OnlineSpeakerDiarization(PipelineConfig.from_namespace(args), profile=True))


if __name__ == "__main__":
    run()



from typing import Text

import numpy as np
from pyannote.core import Annotation, Segment, SlidingWindowFeature


class Binarize:
    """
    Transform a speaker segmentation from the discrete-time domain
    into a continuous-time speaker segmentation.

    Parameters
    ----------
    uri: Text
        Uri of the audio stream.
    threshold: float
        Probability threshold to determine if a speaker is active at a given frame.
    """

    def __init__(self, uri: Text, threshold: float):
        self.uri = uri
        self.threshold = threshold

    def __call__(self, segmentation: SlidingWindowFeature) -> Annotation:
        """
        Return the continuous-time segmentation
        corresponding to the discrete-time input segmentation.

        Parameters
        ----------
        segmentation: SlidingWindowFeature
            Discrete-time speaker segmentation.

        Returns
        -------
        annotation: Annotation
            Continuous-time speaker segmentation.
        """
        num_frames, num_speakers = segmentation.data.shape
        timestamps = segmentation.sliding_window
        is_active = segmentation.data > self.threshold
        # Artificially add last inactive frame to close any remaining speaker turns
        is_active = np.append(is_active, [[False] * num_speakers], axis=0)
        start_times = np.zeros(num_speakers) + timestamps[0].middle
        annotation = Annotation(uri=self.uri, modality="speech")
        for t in range(num_frames):
            # Any (False, True) starts a speaker turn at "True" index
            onsets = np.logical_and(np.logical_not(is_active[t]), is_active[t + 1])
            start_times[onsets] = timestamps[t + 1].middle
            # Any (True, False) ends a speaker turn at "False" index
            offsets = np.logical_and(is_active[t], np.logical_not(is_active[t + 1]))
            for spk in np.where(offsets)[0]:
                region = Segment(start_times[spk], timestamps[t + 1].middle)
                annotation[region, spk] = f"speaker{spk}"
        return annotation



from pathlib import Path
from typing import Union, Text, Optional

import pandas as pd
import rx.operators as ops
from pyannote.core import Annotation
from pyannote.database.util import load_rttm
from pyannote.metrics.diarization import DiarizationErrorRate

import diart.operators as dops
import diart.sources as src
from diart.pipelines import OnlineSpeakerDiarization
from diart.sinks import RTTMWriter, RealTimePlot


class RealTimeInference:
    """
    Streams an audio source to an online speaker diarization pipeline.
    It writes predictions to an output directory in RTTM format and plots them in real time.

    Parameters
    ----------
    output_path: Text or Path
        Output directory to store predictions in RTTM format.
    do_plot: bool
        Whether to draw predictions in a moving plot. Defaults to True.
    """
    def __init__(self, output_path: Union[Text, Path], do_plot: bool = True):
        self.output_path = Path(output_path).expanduser()
        self.output_path.mkdir(parents=True, exist_ok=True)
        self.do_plot = do_plot

    def __call__(self, pipeline: OnlineSpeakerDiarization, source: src.AudioSource) -> Annotation:
        """
        Stream audio chunks from `source` to `pipeline` and write predictions to disk.

        Parameters
        ----------
        pipeline: OnlineSpeakerDiarization
            Configured speaker diarization pipeline.
        source: AudioSource
            Audio source to be read and streamed.

        Returns
        -------
        predictions: Annotation
            Speaker diarization pipeline predictions
        """
        rttm_path = self.output_path / f"{source.uri}.rttm"
        rttm_writer = RTTMWriter(path=rttm_path)
        observable = pipeline.from_audio_source(source).pipe(
            dops.progress(f"Streaming {source.uri}", total=source.length, leave=True)
        )
        if not self.do_plot:
            # Write RTTM file only
            observable.subscribe(rttm_writer)
        else:
            # Write RTTM file + buffering and real-time plot
            observable.pipe(
                ops.do(rttm_writer),
                dops.buffer_output(
                    duration=pipeline.config.duration,
                    step=pipeline.config.step,
                    latency=pipeline.config.latency,
                    sample_rate=pipeline.config.sample_rate
                ),
            ).subscribe(RealTimePlot(pipeline.config.duration, pipeline.config.latency))
        # Stream audio through the pipeline
        source.read()

        return load_rttm(rttm_path)[source.uri]


class Benchmark:
    """
    Run an online speaker diarization pipeline on a set of audio files in batches.
    It writes predictions to a given output directory.

    If the reference is given, it calculates the average diarization error rate.

    Parameters
    ----------
    speech_path: Text or Path
        Directory with audio files.
    reference_path: Text or Path
        Directory with reference RTTM files (same names as audio files).
    output_path: Text or Path
        Output directory to store predictions in RTTM format.
    """
    def __init__(
        self,
        speech_path: Union[Text, Path],
        reference_path: Optional[Union[Text, Path]] = None,
        output_path: Optional[Union[Text, Path]] = None,
        show_progress: bool = True,
        show_report: bool = True,
        batch_size: int = 32,
    ):
        self.speech_path = Path(speech_path).expanduser()
        assert self.speech_path.is_dir(), "Speech path must be a directory"

        self.reference_path = reference_path
        if reference_path is not None:
            self.reference_path = Path(self.reference_path).expanduser()
            assert self.reference_path.is_dir(), "Reference path must be a directory"

        if output_path is None:
            self.output_path = self.speech_path
        else:
            self.output_path = Path(output_path).expanduser()
            self.output_path.mkdir(parents=True, exist_ok=True)

        self.show_progress = show_progress
        self.show_report = show_report
        self.batch_size = batch_size

    def __call__(self, pipeline: OnlineSpeakerDiarization) -> Optional[pd.DataFrame]:
        """
        Run a given pipeline on a set of audio files using
        pre-calculated segmentation and embeddings in batches.

        Parameters
        ----------
        pipeline: OnlineSpeakerDiarization
            Configured speaker diarization pipeline.

        Returns
        -------
        performance: pandas.DataFrame, optional
            DataFrame with detailed performance on each file, as well as average performance.
            None if the reference is not provided.
        """
        loader = src.AudioLoader(pipeline.config.sample_rate, mono=True)
        audio_file_paths = list(self.speech_path.iterdir())
        num_audio_files = len(audio_file_paths)
        for i, filepath in enumerate(audio_file_paths):
            num_chunks = loader.get_num_sliding_chunks(
                filepath, pipeline.config.duration, pipeline.config.step
            )

            # Stream fully online if batch size is 1 or lower
            if self.batch_size < 2:
                source = src.FileAudioSource(
                    filepath,
                    pipeline.config.sample_rate,
                    pipeline.config.duration,
                    pipeline.config.step,
                )
                observable = pipeline.from_audio_source(source)
            else:
                msg = f"Pre-calculating {filepath.stem} ({i + 1}/{num_audio_files})"
                source = src.PrecalculatedFeaturesAudioSource(
                    filepath,
                    pipeline.config.sample_rate,
                    pipeline.segmentation,
                    pipeline.embedding,
                    pipeline.config.duration,
                    pipeline.config.step,
                    self.batch_size,
                    progress_msg=msg if self.show_progress else None,
                )
                observable = pipeline.from_feature_source(source)

            if self.show_progress:
                observable = observable.pipe(
                    dops.progress(
                        desc=f"Streaming {filepath.stem} ({i + 1}/{num_audio_files})",
                        total=num_chunks,
                        leave=False,
                    )
                )

            observable.subscribe(RTTMWriter(path=self.output_path / f"{filepath.stem}.rttm"))

            source.read()

        # Run evaluation
        if self.reference_path is not None:
            metric = DiarizationErrorRate(collar=0, skip_overlap=False)
            for ref_path in self.reference_path.iterdir():
                ref = load_rttm(ref_path).popitem()[1]
                hyp = load_rttm(self.output_path / ref_path.name).popitem()[1]
                metric(ref, hyp)

            return metric.report(display=self.show_report)



from pathlib import Path
from typing import Text, Union

import numpy as np
import torch
import torchaudio
from einops import rearrange
from torchaudio.functional import resample

torchaudio.set_audio_backend("soundfile")


FilePath = Union[Text, Path]


class AudioLoader:
    def __init__(self, sample_rate: int, mono: bool = True):
        self.sample_rate = sample_rate
        self.mono = mono

    def load(self, filepath: FilePath) -> torch.Tensor:
        """Load an audio file into a torch.Tensor.

        Parameters
        ----------
        filepath : FilePath
            Path to an audio file

        Returns
        -------
        waveform : torch.Tensor, shape (channels, samples)
        """
        waveform, sample_rate = torchaudio.load(filepath)
        # Get channel mean if mono
        if self.mono and waveform.shape[0] > 1:
            waveform = waveform.mean(dim=0, keepdim=True)
        # Resample if needed
        if self.sample_rate != sample_rate:
            waveform = resample(waveform, sample_rate, self.sample_rate)
        return waveform

    @staticmethod
    def get_duration(filepath: FilePath) -> float:
        """Get audio file duration in seconds.

        Parameters
        ----------
        filepath : FilePath
            Path to an audio file

        Returns
        -------
        duration : float
            Duration in seconds.
        """
        info = torchaudio.info(filepath)
        return info.num_frames / info.sample_rate

    def load_sliding_chunks(self, filepath: FilePath, chunk_duration: float, step_duration: float) -> np.ndarray:
        """Load an audio file and extract sliding chunks of a given duration with a given step duration.

        Parameters
        ----------
        filepath : FilePath
            Path to an audio file
        chunk_duration: float
            Duration of the chunk in seconds.
        step_duration: float
            Duration of the step between chunks in seconds.
        """
        chunk_samples = int(round(chunk_duration * self.sample_rate))
        step_samples = int(round(step_duration * self.sample_rate))
        waveform = self.load(filepath)
        _, num_samples = waveform.shape
        chunks = rearrange(
            waveform.unfold(1, chunk_samples, step_samples),
            "channel chunk sample -> chunk channel sample",
        ).numpy()
        # Add padded last chunk
        if num_samples - chunk_samples % step_samples > 0:
            last_chunk = waveform[:, chunks.shape[0] * step_samples:].unsqueeze(0).numpy()
            diff_samples = chunk_samples - last_chunk.shape[-1]
            last_chunk = np.concatenate([last_chunk, np.zeros((1, 1, diff_samples))], axis=-1)
            return np.vstack([chunks, last_chunk])
        return chunks

    def get_num_sliding_chunks(self, filepath: FilePath, chunk_duration: float, step_duration: float) -> int:
        """Estimate the number of sliding chunks of a
        given chunk duration and step without loading the audio.

        Parameters
        ----------
        filepath : FilePath
            Path to an audio file
        chunk_duration: float
            Duration of the chunk in seconds.
        step_duration: float
            Duration of the step between chunks in seconds.
        """
        numerator = self.get_duration(filepath) - chunk_duration + step_duration
        return int(np.ceil(numerator / step_duration))



from typing import Optional, List

import numpy as np
from pyannote.core import Segment, SlidingWindow, SlidingWindowFeature
from typing_extensions import Literal


class AggregationStrategy:
    """Abstract class representing a strategy to aggregate overlapping buffers"""

    @staticmethod
    def build(name: Literal["mean", "hamming", "first"]) -> 'AggregationStrategy':
        """Build an AggregationStrategy instance based on its name"""
        assert name in ("mean", "hamming", "first")
        if name == "mean":
            return AverageStrategy()
        elif name == "hamming":
            return HammingWeightedAverageStrategy()
        else:
            return FirstOnlyStrategy()

    def __call__(self, buffers: List[SlidingWindowFeature], focus: Segment) -> SlidingWindowFeature:
        """Aggregate chunks over a specific region.

        Parameters
        ----------
        buffers: list of SlidingWindowFeature, shapes (frames, speakers)
            Buffers to aggregate
        focus: Segment
            Region to aggregate that is shared among the buffers

        Returns
        -------
        aggregation: SlidingWindowFeature, shape (cropped_frames, speakers)
            Aggregated values over the focus region
        """
        aggregation = self.aggregate(buffers, focus)
        resolution = focus.duration / aggregation.shape[0]
        resolution = SlidingWindow(
            start=focus.start,
            duration=resolution,
            step=resolution
        )
        return SlidingWindowFeature(aggregation, resolution)

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        raise NotImplementedError


class HammingWeightedAverageStrategy(AggregationStrategy):
    """Compute the average weighted by the corresponding Hamming-window aligned to each buffer"""

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        num_frames, num_speakers = buffers[0].data.shape
        hamming, intersection = [], []
        for buffer in buffers:
            # Crop buffer to focus region
            b = buffer.crop(focus, fixed=focus.duration)
            # Crop Hamming window to focus region
            h = np.expand_dims(np.hamming(num_frames), axis=-1)
            h = SlidingWindowFeature(h, buffer.sliding_window)
            h = h.crop(focus, fixed=focus.duration)
            hamming.append(h.data)
            intersection.append(b.data)
        hamming, intersection = np.stack(hamming), np.stack(intersection)
        # Calculate weighted mean
        return np.sum(hamming * intersection, axis=0) / np.sum(hamming, axis=0)


class AverageStrategy(AggregationStrategy):
    """Compute a simple average over the focus region"""

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        # Stack all overlapping regions
        intersection = np.stack([
            buffer.crop(focus, fixed=focus.duration)
            for buffer in buffers
        ])
        return np.mean(intersection, axis=0)


class FirstOnlyStrategy(AggregationStrategy):
    """Instead of aggregating, keep the first focus region in the buffer list"""

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        return buffers[0].crop(focus, fixed=focus.duration)


class DelayedAggregation:
    """Aggregate aligned overlapping windows of the same duration
    across sliding buffers with a specific step and latency.

    Parameters
    ----------
    step: float
        Shift between two consecutive buffers, in seconds.
    latency: float, optional
        Desired latency, in seconds. Defaults to step.
        The higher the latency, the more overlapping windows to aggregate.
    strategy: ("mean", "hamming", "any"), optional
        Specifies how to aggregate overlapping windows. Defaults to "hamming".
        "mean": simple average
        "hamming": average weighted by the Hamming window values (aligned to the buffer)
        "any": no aggregation, pick the first overlapping window
    stream_end: float, optional
        Stream end time (in seconds). Defaults to None.
        If the stream end time is known, then append remaining outputs at the end,
        otherwise the last `latency - step` seconds are ignored.

    Example
    --------
    >>> duration = 5
    >>> frames = 500
    >>> step = 0.5
    >>> speakers = 2
    >>> start_time = 10
    >>> resolution = duration / frames
    >>> dagg = DelayedAggregation(step=step, latency=2, strategy="mean")
    >>> buffers = [
    >>>     SlidingWindowFeature(
    >>>         np.random.rand(frames, speakers),
    >>>         SlidingWindow(start=(i + start_time) * step, duration=resolution, step=resolution)
    >>>     )
    >>>     for i in range(dagg.num_overlapping_windows)
    >>> ]
    >>> dagg.num_overlapping_windows
    ... 4
    >>> dagg(buffers).data.shape
    ... (51, 2)  # Rounding errors are possible when cropping the buffers
    """

    def __init__(
        self,
        step: float,
        latency: Optional[float] = None,
        strategy: Literal["mean", "hamming", "first"] = "hamming",
        stream_end: Optional[float] = None
    ):
        self.step = step
        self.latency = latency
        self.strategy = strategy
        self.stream_end = stream_end

        if self.latency is None:
            self.latency = self.step

        assert self.step <= self.latency, "Invalid latency requested"

        self.num_overlapping_windows = int(round(self.latency / self.step))
        self.aggregate = AggregationStrategy.build(self.strategy)

    def _prepend_or_append(
        self,
        output_window: SlidingWindowFeature,
        output_region: Segment,
        buffers: List[SlidingWindowFeature]
    ):
        last_buffer = buffers[-1].extent
        # Prepend prediction until we match the latency in case of first buffer
        if len(buffers) == 1 and last_buffer.start == 0:
            num_frames = output_window.data.shape[0]
            first_region = Segment(0, output_region.end)
            first_output = buffers[0].crop(
                first_region, fixed=first_region.duration
            )
            first_output[-num_frames:] = output_window.data
            resolution = output_region.end / first_output.shape[0]
            output_window = SlidingWindowFeature(
                first_output,
                SlidingWindow(start=0, duration=resolution, step=resolution)
            )
        # Append rest of the outputs
        elif self.stream_end is not None and last_buffer.end == self.stream_end:
            # FIXME instead of appending a larger chunk than expected when latency > step,
            #  keep emitting windows until the signal ends.
            #  This should be fixed at the observable level and not within the aggregation block.
            num_frames = output_window.data.shape[0]
            last_region = Segment(output_region.start, last_buffer.end)
            last_output = buffers[-1].crop(
                last_region, fixed=last_region.duration
            )
            last_output[:num_frames] = output_window.data
            resolution = self.latency / last_output.shape[0]
            output_window = SlidingWindowFeature(
                last_output,
                SlidingWindow(
                    start=output_region.start,
                    duration=resolution,
                    step=resolution
                )
            )
        return output_window

    def __call__(self, buffers: List[SlidingWindowFeature]) -> SlidingWindowFeature:
        # Determine overlapping region to aggregate
        start = buffers[-1].extent.end - self.latency
        region = Segment(start, start + self.step)
        return self._prepend_or_append(self.aggregate(buffers, region), region, buffers)



import math
from pathlib import Path
from queue import SimpleQueue
from typing import Text, Optional, Callable

import numpy as np
import sounddevice as sd
import torch
from einops import rearrange
from pyannote.core import SlidingWindowFeature, SlidingWindow
from rx.subject import Subject
from tqdm import tqdm

from .audio import FilePath, AudioLoader
from .features import TemporalFeatures


class AudioSource:
    """Represents a source of audio that can start streaming via the `stream` property.

    Parameters
    ----------
    uri: Text
        Unique identifier of the audio source.
    sample_rate: int
        Sample rate of the audio source.
    """
    def __init__(self, uri: Text, sample_rate: int):
        self.uri = uri
        self.sample_rate = sample_rate
        self.stream = Subject()

    @property
    def is_regular(self) -> bool:
        """Whether the stream is regular. Defaults to False.
        A regular stream always yields the same amount of samples per event.
        """
        return False

    @property
    def duration(self) -> Optional[float]:
        """The duration of the stream if known. Defaults to None (unknown duration)"""
        return None

    @property
    def length(self) -> Optional[int]:
        """Return the number of audio chunks emitted by this source"""
        return None

    def read(self):
        """Start reading the source and yielding samples through the stream"""
        raise NotImplementedError


class FileAudioSource(AudioSource):
    """Represents an audio source tied to a file.

    Parameters
    ----------
    file: FilePath
        Path to the file to stream.
    sample_rate: int
        Sample rate of the chunks emitted.
    chunk_duration: float
        Duration of each chunk in seconds. Defaults to 5s.
    step_duration: float
        Duration of the step between consecutive chunks in seconds. Defaults to 500ms.
    """
    def __init__(
        self,
        file: FilePath,
        sample_rate: int,
        chunk_duration: float = 5,
        step_duration: float = 0.5,
    ):
        super().__init__(Path(file).stem, sample_rate)
        self.loader = AudioLoader(sample_rate, mono=True)
        self._duration = self.loader.get_duration(file)
        self.file = file
        self.chunk_duration = chunk_duration
        self.step_duration = step_duration
        self.resolution = 1 / sample_rate

    @property
    def is_regular(self) -> bool:
        # An audio file is always a regular source
        return True

    @property
    def duration(self) -> Optional[float]:
        # The duration of a file is known
        return self._duration

    @property
    def length(self) -> Optional[int]:
        return self.loader.get_num_sliding_chunks(
            self.file, self.chunk_duration, self.step_duration
        )

    def read(self):
        """Send each chunk of samples through the stream"""
        chunks = self.loader.load_sliding_chunks(
            self.file, self.chunk_duration, self.step_duration
        )
        for i, waveform in enumerate(chunks):
            window = SlidingWindow(
                start=i * self.step_duration,
                duration=self.resolution,
                step=self.resolution
            )
            chunk = SlidingWindowFeature(waveform.T, window)
            try:
                self.stream.on_next(chunk)
            except Exception as e:
                self.stream.on_error(e)
        self.stream.on_completed()


class PrecalculatedFeaturesAudioSource(FileAudioSource):
    def __init__(
        self,
        file: FilePath,
        sample_rate: int,
        segmentation: Callable[[TemporalFeatures], TemporalFeatures],
        embedding: Callable[[TemporalFeatures, TemporalFeatures], TemporalFeatures],
        chunk_duration: float = 5,
        step_duration: float = 0.5,
        batch_size: int = 32,
        progress_msg: Optional[Text] = None,
    ):
        super().__init__(file, sample_rate, chunk_duration, step_duration)
        self.segmentation = segmentation
        self.embedding = embedding
        self.batch_size = batch_size
        self.progress_msg = progress_msg

    def read(self):
        # Split audio into chunks
        chunks = rearrange(
            self.loader.load_sliding_chunks(
                self.file, self.chunk_duration, self.step_duration
            ),
            "chunk channel sample -> chunk sample channel"
        )
        num_chunks = chunks.shape[0]

        # Set progress if needed
        iterator = range(0, num_chunks, self.batch_size)
        if self.progress_msg is not None:
            total = int(math.ceil(num_chunks / self.batch_size))
            iterator = tqdm(iterator, desc=self.progress_msg, total=total, unit="batch", leave=False)

        # Pre-calculate segmentation and embeddings
        segmentation, embeddings = [], []
        for i in iterator:
            i_end = i + self.batch_size
            if i_end > num_chunks:
                i_end = num_chunks
            batch = chunks[i:i_end]
            seg = self.segmentation(batch)
            # Edge case: add batch dimension if i == i_end + 1
            if seg.ndim == 2:
                seg = seg[np.newaxis]
            emb = self.embedding(batch, seg)
            # Edge case: add batch dimension if i == i_end + 1
            if emb.ndim == 2:
                emb = emb.unsqueeze(0)
            segmentation.append(seg)
            embeddings.append(emb)
        segmentation = np.vstack(segmentation)
        embeddings = torch.vstack(embeddings)

        # Stream pre-calculated segmentation, embeddings and chunks
        seg_resolution = self.chunk_duration / segmentation.shape[1]
        for i in range(num_chunks):
            chunk_window = SlidingWindow(
                start=i * self.step_duration,
                duration=self.resolution,
                step=self.resolution,
            )
            seg_window = SlidingWindow(
                start=i * self.step_duration,
                duration=seg_resolution,
                step=seg_resolution,
            )
            try:
                self.stream.on_next((
                    SlidingWindowFeature(chunks[i], chunk_window),
                    SlidingWindowFeature(segmentation[i], seg_window),
                    embeddings[i]
                ))
            except Exception as e:
                self.stream.on_error(e)

        self.stream.on_completed()


class MicrophoneAudioSource(AudioSource):
    """Represents an audio source tied to the default microphone available"""

    def __init__(self, sample_rate: int):
        super().__init__("live_recording", sample_rate)
        self.block_size = 1024
        self.mic_stream = sd.InputStream(
            channels=1,
            samplerate=sample_rate,
            latency=0,
            blocksize=self.block_size,
            callback=self._read_callback
        )
        self.queue = SimpleQueue()

    def _read_callback(self, samples, *args):
        self.queue.put_nowait(samples[:, [0]].T)

    def read(self):
        self.mic_stream.start()
        while self.mic_stream:
            try:
                self.stream.on_next(self.queue.get())
            except Exception as e:
                self.stream.on_error(e)
                break
        self.stream.on_completed()



from typing import Optional, List, Any

import rx
import rx.operators as ops
import torch

from . import blocks
from . import models as m
from . import operators as dops
from . import sources as src
from . import utils


class PipelineConfig:
    def __init__(
        self,
        segmentation: Optional[m.SegmentationModel] = None,
        embedding: Optional[m.EmbeddingModel] = None,
        duration: Optional[float] = None,
        step: float = 0.5,
        latency: Optional[float] = None,
        tau_active: float = 0.6,
        rho_update: float = 0.3,
        delta_new: float = 1,
        gamma: float = 3,
        beta: float = 10,
        max_speakers: int = 20,
        device: Optional[torch.device] = None,
        **kwargs,
    ):
        # Default segmentation model is pyannote/segmentation
        self.segmentation = segmentation
        if self.segmentation is None:
            self.segmentation = m.SegmentationModel.from_pyannote("pyannote/segmentation")

        # Default duration is the one given by the segmentation model
        self.duration = duration
        if self.duration is None:
            self.duration = self.segmentation.get_duration()

        # Expected sample rate is given by the segmentation model
        self.sample_rate = self.segmentation.get_sample_rate()

        # Default embedding model is pyannote/embedding
        self.embedding = embedding
        if self.embedding is None:
            self.embedding = m.EmbeddingModel.from_pyannote("pyannote/embedding")

        # Latency defaults to the step duration
        self.step = step
        self.latency = latency
        if self.latency is None:
            self.latency = self.step

        self.tau_active = tau_active
        self.rho_update = rho_update
        self.delta_new = delta_new
        self.gamma = gamma
        self.beta = beta
        self.max_speakers = max_speakers

        self.device = device
        if self.device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    @staticmethod
    def from_namespace(args: Any) -> 'PipelineConfig':
        return PipelineConfig(
            segmentation=getattr(args, "segmentation", None),
            embedding=getattr(args, "embedding", None),
            duration=getattr(args, "duration", None),
            step=args.step,
            latency=args.latency,
            tau_active=args.tau,
            rho_update=args.rho,
            delta_new=args.delta,
            gamma=args.gamma,
            beta=args.beta,
            max_speakers=args.max_speakers,
            device=args.device,
        )

    def last_chunk_end_time(self, conv_duration: float) -> Optional[float]:
        """Return the end time of the last chunk for a given conversation duration.

        Parameters
        ----------
        conv_duration: float
            Duration of a conversation in seconds.
        """
        return conv_duration - conv_duration % self.step


class OnlineSpeakerTracking:
    def __init__(self, config: PipelineConfig):
        self.config = config

    def get_end_time(self, duration: Optional[float]) -> Optional[float]:
        return None if duration is None else self.config.last_chunk_end_time(duration)

    def get_operators(self, source: src.AudioSource) -> List[dops.Operator]:
        clustering = blocks.OnlineSpeakerClustering(
            self.config.tau_active,
            self.config.rho_update,
            self.config.delta_new,
            "cosine",
            self.config.max_speakers,
        )
        end_time = self.get_end_time(source.duration)
        pred_aggregation = blocks.DelayedAggregation(
            self.config.step, self.config.latency, strategy="hamming", stream_end=end_time
        )
        audio_aggregation = blocks.DelayedAggregation(
            self.config.step, self.config.latency, strategy="first", stream_end=end_time
        )
        binarize = blocks.Binarize(source.uri, self.config.tau_active)
        return [
            # Identify global speakers with online clustering
            ops.starmap(lambda wav, seg, emb: (wav, clustering(seg, emb))),
            # Buffer 'num_overlapping' sliding chunks with a step of 1 chunk
            dops.buffer_slide(pred_aggregation.num_overlapping_windows),
            # Aggregate overlapping output windows
            ops.map(utils.unzip),
            ops.starmap(lambda wav_buffer, pred_buffer: (
                audio_aggregation(wav_buffer), pred_aggregation(pred_buffer)
            )),
            # Binarize output
            ops.starmap(lambda wav, pred: (binarize(pred), wav)),
        ]


class OnlineSpeakerDiarization:
    def __init__(self, config: PipelineConfig, profile: bool = False):
        self.config = config
        self.profile = profile
        self.segmentation = blocks.SpeakerSegmentation(config.segmentation, config.device)
        self.embedding = blocks.OverlapAwareSpeakerEmbedding(
            config.embedding, config.gamma, config.beta, norm=1, device=config.device
        )
        self.speaker_tracking = OnlineSpeakerTracking(config)
        msg = f"Latency should be in the range [{config.step}, {config.duration}]"
        assert config.step <= config.latency <= config.duration, msg

    def from_audio_source(self, source: src.AudioSource) -> rx.Observable:
        msg = f"Audio source has sample rate {source.sample_rate}, expected {self.config.sample_rate}"
        assert source.sample_rate == self.config.sample_rate, msg
        operators = []
        # Regularize the stream to a specific chunk duration and step
        if not source.is_regular:
            operators.append(dops.regularize_audio_stream(
                self.config.duration, self.config.step, source.sample_rate
            ))
        operators += [
            # Extract segmentation and keep audio
            ops.map(lambda wav: (wav, self.segmentation(wav))),
            # Extract embeddings and keep segmentation
            ops.starmap(lambda wav, seg: (wav, seg, self.embedding(wav, seg))),
        ]
        # Add speaker tracking
        operators += self.speaker_tracking.get_operators(source)
        if self.profile:
            return dops.profile(source.stream, operators)
        return source.stream.pipe(*operators)

    def from_feature_source(self, source: src.AudioSource) -> rx.Observable:
        return source.stream.pipe(*self.speaker_tracking.get_operators(source))



from typing import Optional

import torch
from einops import rearrange

from ..features import TemporalFeatures, TemporalFeatureFormatter
from ..models import SegmentationModel


class SpeakerSegmentation:
    def __init__(self, model: SegmentationModel, device: Optional[torch.device] = None):
        self.model = model
        self.model.eval()
        self.device = device
        if self.device is None:
            self.device = torch.device("cpu")
        self.model.to(self.device)
        self.formatter = TemporalFeatureFormatter()

    @staticmethod
    def from_pyannote(model, device: Optional[torch.device] = None) -> 'SpeakerSegmentation':
        return SpeakerSegmentation(SegmentationModel.from_pyannote(model), device)

    def __call__(self, waveform: TemporalFeatures) -> TemporalFeatures:
        """
        Calculate the speaker segmentation of input audio.

        Parameters
        ----------
        waveform: TemporalFeatures, shape (samples, channels) or (batch, samples, channels)

        Returns
        -------
        speaker_segmentation: TemporalFeatures, shape (batch, frames, speakers)
            The batch dimension is omitted if waveform is a `SlidingWindowFeature`.
        """
        with torch.no_grad():
            wave = rearrange(self.formatter.cast(waveform), "batch sample channel -> batch channel sample")
            output = self.model(wave.to(self.device)).cpu()
        return self.formatter.restore_type(output)


from typing import Optional

import torch
import torch.nn as nn

try:
    import pyannote.audio.pipelines.utils as pyannote
    _has_pyannote = True
except ImportError:
    _has_pyannote = False


class SegmentationModel(nn.Module):
    """
    Minimal interface for a segmentation model.
    """

    @staticmethod
    def from_pyannote(model) -> 'SegmentationModel':
        """
        Returns a `SegmentationModel` wrapping a pyannote model.

        Parameters
        ----------
        model: pyannote.PipelineModel

        Returns
        -------
        wrapper: SegmentationModel
        """
        assert _has_pyannote, "No pyannote.audio installation found"

        class PyannoteSegmentationModel(SegmentationModel):
            def __init__(self, pyannote_model):
                super().__init__()
                self.model = pyannote.get_model(pyannote_model)

            def get_sample_rate(self) -> int:
                return self.model.audio.sample_rate

            def get_duration(self) -> float:
                return self.model.specifications.duration

            def __call__(self, waveform: torch.Tensor) -> torch.Tensor:
                return self.model(waveform)

        return PyannoteSegmentationModel(model)

    def get_sample_rate(self) -> int:
        """Return the sample rate expected for model inputs"""
        raise NotImplementedError

    def get_duration(self) -> float:
        """Return the input duration by default (usually the one used during training)"""
        raise NotImplementedError

    def __call__(self, waveform: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of a segmentation model.

        Parameters
        ----------
        waveform: torch.Tensor, shape (batch, channels, samples)

        Returns
        -------
        speaker_segmentation: torch.Tensor, shape (batch, frames, speakers)
        """
        raise NotImplementedError


class EmbeddingModel(nn.Module):
    """Minimal interface for an embedding model."""

    @staticmethod
    def from_pyannote(model) -> 'EmbeddingModel':
        """
        Returns an `EmbeddingModel` wrapping a pyannote model.

        Parameters
        ----------
        model: pyannote.PipelineModel

        Returns
        -------
        wrapper: EmbeddingModel
        """
        assert _has_pyannote, "No pyannote.audio installation found"

        class PyannoteEmbeddingModel(EmbeddingModel):
            def __init__(self, pyannote_model):
                super().__init__()
                self.model = pyannote.get_model(pyannote_model)

            def __call__(
                self,
                waveform: torch.Tensor,
                weights: Optional[torch.Tensor] = None,
            ) -> torch.Tensor:
                return self.model(waveform, weights=weights)

        return PyannoteEmbeddingModel(model)

    def __call__(
        self,
        waveform: torch.Tensor,
        weights: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Forward pass of an embedding model with optional weights.

        Parameters
        ----------
        waveform: torch.Tensor, shape (batch, channels, samples)
        weights: Optional[torch.Tensor], shape (batch, frames)
            Temporal weights for each sample in the batch. Defaults to no weights.

        Returns
        -------
        speaker_embeddings: torch.Tensor, shape (batch, speakers, embedding_dim)
        """
        raise NotImplementedError
